#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "aa.h"
#include "mem.h"

/* NOTE: the functions that take or return an atomic number
   use the standard 1 = H, etc. Internally (for common elements),
   we work with 92 - N, so that 0 corresponds to U and 91 to H.
   In the comments below, 'atomic number' means this version.
*/

/* A 'numbered element'; ie., an element and its atomic number. */
struct nbdelt {
    struct cmpd *c;
    int an;
};

/* The maximum number of daughters of a common element. */
#define MAXDTRS 6

/* A table of the common elements and their atomic numbers,
   in lexicograpic order of the digit strings.
   Filled in by init_periodic_table().
*/
static struct nbdelt table[U];

/* lkup[N] is the index of element N in table[].
   Filled in by init_periodic_table().
*/
static int lkup[U];

/* dtrs[N][0] is the number D of daughters of element N; the atomic
   numbers of the daughters are in dtrs[N][1] to dtrs[N][D].
   Filled in by init_periodic_table().
*/
static int dtrs[U][MAXDTRS + 1];

/* The chemical symbols for the common elements. */
static const char *names[U] =
#include "elts.txt"
;

void killcmpd(struct cmpd *c)
{
    while (c) {
        struct cmpd *d = c;
        c = c->next;
        xfree(d);
    }
}

void killlist(struct list *p)
{
    while (p) {
        struct list *q = p;
        p = p->next;
        killcmpd(q->c);
        xfree(q);
    }
}

struct cmpd *readcmpd(const char *s)
{
    size_t i;
    struct cmpd *ret = NULL, *t;

    i = strspn(s, "123456789");
    while (i--) {
        t = xalloc(sizeof *t);
        t->n = s[i] - '0';
        t->next = ret;
        ret = t;
    }
    return ret;
}
    
void fprtcmpd(const struct cmpd *c, FILE *f)
{
    while (c) {
        if (c->n <= 9)
            fputc(c->n + '0', f);
        else
            fprintf(f, "{%d}", c->n);
        c = c->next;
    }
}

void fprtlist_dots(const struct list *p, FILE *f)
{
    int first = 1;

    while (p) {
        if (first)
            first = 0;
        else
            fputc('.', f);
        fprtcmpd(p->c, f);
        p = p->next;
    }
}

void fprtlist_nls(const struct list *p, FILE *f)
{
    while (p) {
        fputcmpd(p->c, f);
        p = p->next;
    }
}

int cmpcmpd(const struct cmpd *a, const struct cmpd *b)
{
    while (a && b && a->n == b->n) {
        a = a->next;
        b = b->next;
    }
    if (!a)
        return !b ? 0 : -1;
    return !b || b->n < a->n ? 1 : -1;
}

struct cmpd *cpycmpd(const struct cmpd *c)
{
    struct cmpd head, *last;

    last = &head;
    while (c) {
        last->next = xalloc(sizeof *last);
        last = last->next;
        last->n = c->n;
        c = c->next;
    }
    last->next = NULL;
    return head.next;
}

struct cmpd *dtr(struct cmpd *c)
{
    struct cmpd head, *e;

    e = &head;
    while (c) {
        int n = c->n, ct = 1;
        while ((c = c->next) && c->n == n)
            ++ct;
        e->next = xalloc(sizeof *e);
        e = e->next;
        e->n = ct;
        e->next = xalloc(sizeof *e);
        e = e->next;
        e->n = n;
    }
    e->next = NULL;
    return head.next;
}

int cansplit(struct cmpd *c, int n)
{
    struct cmpd h;
    int ret;
    
    if (n >= 2)
        return cansplit2(c);
    if (!c || !c->next || c->n == c->next->n)
        return 0;
    h.n = c->n;
    h.next = dtr(c->next);
    ret = cansplit(&h, n+1);
    killcmpd(h.next);
    return ret;
}

int cansplit2(struct cmpd *c)
{
    struct cmpd *d;
    
    if (!c || !(d = c->next))
        return 0;
    if (c->n > 3)
        return 1;
    if (c->n != 2) {
        if (d->n != 2 || !d->next)
            return 0;
        d = d->next;
        if (d->n != 2)
            return 0;
        d = d->next;
        if (!d)
            return 1;
    }
    switch (d->n) {
    case 1:
        d = d->next;
        if (!d)
            return 0;
        if (d->n == 1) {
            d = d->next;
            return (d && d->n == 1);
        }
        else
            return !d->next || d->next->n != d->n;
    case 2:
        return 0;
    case 3:
        d = d->next;
        if (!d)
            return 1;
        return d->n != 3 && (!d->next || d->next->n != d->n
                             || !d->next->next || d->next->next->n != d->n);
    default:
        return 1;
    }
}

struct list *split(struct cmpd *c, int n)
{
    struct list *ret, *e;

    if (!c)
        return NULL;
    ret = e = xalloc(sizeof *e);
    e->c = c;
    do {
        if (cansplit(c, n)) {
            e->next = xalloc(sizeof *e);
            e = e->next;
            e->c = c->next;
            c->next = NULL;
            c = e->c;
        }
        else
            c = c->next;
    } while (c);
    e->next = NULL;
    return ret;
}

/* cmp_nbdelt(A, B): A and B are really of type 'struct nbdelt *',
   the declaration matches the comparison function argument to
   qsort() and bsearch().
*/
static int cmp_nbdelt(const void *a, const void *b)
{
    return cmpcmpd( ((const struct nbdelt *)a)->c,
                    ((const struct nbdelt *)b)->c);
}

void init_periodic_table(void)
{
    int found, done, i, j, stack[U], nd, dtrs2[U][MAXDTRS + 1];

    /* Generate the common elements and the lists of their daughters
       by repeated decay and splitting from the string 3. We put
       the daughters in the local variable dtrs2[], not dtrs[], to
       avoid having to shuffle them in place later.
    */
    table[0].c = readcmpd("3");
    found = 1;
    done = 0;
    do {
        struct list *p = split(dtr(table[done].c), 2), *q;
        nd = 0;
        do {
            for (i = 0; i < found && cmpcmpd(table[i].c, p->c) != 0; ++i)
                ;
            if (i < found) 
                killcmpd(p->c);
            else {
                assert(found < U);
                table[found++].c = p->c;
            }
            assert(nd < MAXDTRS);
            ++nd;
            dtrs2[done][nd] = i;
            q = p;
            p = p->next;
            xfree(q);
        } while (p);
        dtrs2[done][0] = nd;
        ++done;
    } while (done < found);
    assert(found == U);

    /* Find the atomic numbers of the elements. That the following code
       does this is a bit of a fluke, because Conway's ordering is
       somewhat arbitrary. It has the property that each element except
       the first is among the daughters of its predecessor, but there
       are seven such orderings, four starting with U and three with Sr.
       It happens that using the obvious backtracking algorithm, starting
       from U and trying daughters in left-to-right order, the first
       ordering found is Conway's.
       This piece of code fills in the atomic number fields in table[],
       and also sets lkup[] correctly for the current ordering of table[]. 
    */
    lkup[0] = 0;
    table[0].an = 0;
    for (i = 1; i < U; ++i)
        table[i].an = -1;
    stack[0] = 1;
    i = 0;
    do {
        if (stack[i] > dtrs2[lkup[i]][0]) {
            assert(i > 0);
            table[lkup[i]].an = -1;
            --i;
            ++stack[i];
        }
        else {
            j = dtrs2[lkup[i]][stack[i]];
            if (table[j].an < 0) {
                ++i;
                table[j].an = i;
                lkup[i] = j;
                stack[i] = 1;
            }
            else
                ++stack[i];
        }
    } while (i < U - 1);

    /* Fill in dtrs[] in order of atomic number. */
    for (i = 0; i < U; ++i) {
        dtrs[i][0] = nd = dtrs2[lkup[i]][0];
        for (j = 1; j <= nd; ++j)
            dtrs[i][j] = table[dtrs2[lkup[i]][j]].an;
    }

    /* Sort table[] by lexicographic order of the digit strings. */
    qsort(table, U, sizeof table[0], cmp_nbdelt);

    /* Compute lkup[] for the new ordering of table[]. */
    for (i = 0; i < U; ++i)
        lkup[table[i].an] = i;
}

const struct cmpd *elt_str(int n)
{
    return (1 <= n && n <= U) ? table[lkup[U - n]].c : NULL;
}

int num_dtrs(int n)
{
    return (1 <= n && n <= U) ? dtrs[U - n][0] : 0;
}

int get_dtr(int par, int i)
{
    int  nd = (1 <= par && par <= U) ? dtrs[U - par][0] : 0;
    return (1 <= i && i <= nd) ? U - dtrs[U - par][i] : 0;
}

const char *elt_name(int n)
{
    if (1 <= n && n <= U)
        return names[U - n];
    switch (n) {
    case U + 1:
        return "Np";
    case U + 2:
        return "Pu";
    default:
        return "";
    }
}

int is_common(struct cmpd *c)
{
    struct nbdelt *x, y;
    
    y.c = c;
    x = bsearch(&y, table, U, sizeof y, cmp_nbdelt);
    return x ? U - x->an : 0;
}

int is_transu(struct cmpd *c)
{
    int i;
    struct cmpd *a, *b;

    for (i = 2; i <= 3; ++i) {
        a = table[lkup[U - i]].c;
        b = c;
        while (a && b && a->n == b->n) {
            a = a->next;
            b = b->next;
        }
        if (a && b && !a->next && !b->next && b->n > 3)
            return U + i - 1;
    }
    return 0;
}

int longevity(struct cmpd *c)
{
    struct list *p, *q, *r;
    int ret, done;

    if (!c || is_common(c) || is_transu(c))
        return 0;
    p = split(dtr(c), 2);
    ret = 0;
    do {
        done = 1;
        ++ret;
        q = p;
        do {
            if (!(c = q->c))
                continue;
            q->c = NULL;
            if (!(is_common(c) || is_transu(c))) {
                done = 0;
                r = q->next;
                q->next = split(dtr(c), 2);
                do {
                    q = q->next;
                } while (q->next);
                q->next = r;
            }
            killcmpd(c);
        } while ((q = q->next));
    } while (!done);
    killlist(p);
    return ret;
}

void kill_periodic_table(void)
{
    int i;

    for (i = 0; i < U; ++i)
        killcmpd(table[i].c);
}

