#ifndef AA_H
#define AA_H

#include <stdio.h>

/* Data structures and functions implementing audioactive decay. */

#define U 92

/* Compounds (digit strings) are represented as linked lists. */
struct cmpd {
    int n;
    struct cmpd *next;
};

/* Lists of compounds. */
struct list {
    struct cmpd *c;
    struct list *next;
};

/* killcmpd(C): free the memory used by compound C. */
void killcmpd(struct cmpd *);

/* killlist(P): free the memory used by list P and all the
   compounds in the list.
*/
void killlist(struct list *);

/* readcmpd(S): create the compound determined by the character
   string S (eg. "12113").
*/
struct cmpd *readcmpd(const char *);

/* fprtcmpd(C, F): print the compound C to the file F.
   If C contains 'digits' greater than 9, they are printed
   in the form {10}.
*/
void fprtcmpd(const struct cmpd *, FILE *);

/* fputcmpd(C, F): print the compound C to the file F, followed by
   a newline.
*/
#define fputcmpd(c, f) do {fprtcmpd(c, f); fputc('\n', f);} while (0)

/* fprtlist_dots(P, F): print the compounds in the list P to the
   file F, separated by dots.
*/
void fprtlist_dots(const struct list *, FILE *);

/* fprtlist_nls(P, F): print the compounds in the list P to the
   file F, each followed by a newline.
*/
void fprtlist_nls(const struct list *, FILE *);

/* Macros for the case F == stdout of the last four functions/macros. */
#define prtcmpd(c) fprtcmpd(c, stdout)
#define putcmpd(c) fputcmpd(c, stdout)
#define prtlist_dots(p) fprtlist_dots(p, stdout)
#define prtlist_nls(p) fprtlist_nls(p, stdout)

/* cmpcmpd(A, B): compare the compounds A and B in lexicographic
   order; return negative if A < B, 0 if A == B, and positive
   if A > B.
*/
int cmpcmpd(const struct cmpd *, const struct cmpd *);

/* cpycmpd(C): return a copy of the compound C.*/
struct cmpd *cpycmpd(const struct cmpd *);

/* dtr(C): return the daughter of the compound C; C is unchanged. */
struct cmpd *dtr(struct cmpd *);

/* cansplit(C, N): check if the N-day-old compound C splits after
   its first digit; return non-zero if so, 0 if not. For N >= 2, we
   just use the splitting theorem; the cases where N is 0 or 1 are special.
*/
int cansplit (struct cmpd *, int);

/* cansplit2(C): equivalent to cansplit(C, 2). */
int cansplit2(struct cmpd *);

/* split(C, N): split the N-day-old compound C into its constituent
   elements, and return a list of them (in order). C is destroyed.
   Again, all values of N >= 2 are equivalent; 0 and 1 are special.
*/
struct list *split(struct cmpd *, int);

/* init_periodic_table(): construct the periodic table. Must be
   called before the functions below.
*/
void init_periodic_table(void);

/* elt_str(N): if 1 <= N and N <= 92, return the element of atomic
   number N, otherwise return NULL.
*/
const struct cmpd *elt_str(int);

/* num_dtrs(N): if 1 <= N and N <= 92, return the number of daughters
   of the element of atomic number N, otherwise return 0.
*/
int num_dtrs(int);

/* get_dtr(N, I): if 1 <= I and I <= num_dtrs(N), return the atomic
   number of the Ith daughter of element N, otherwise return 0.
*/
int get_dtr(int, int);

/* elt_name(N): if 1 <= N and N <= 94, return the chemical symbol
   for element N ("H", "He", ..., "Pu"), otherwise return "".
*/
const char *elt_name(int);

/* is_common(C): if the compound C is a common element, return its
   atomic number, otherwise return 0.
*/
int is_common(struct cmpd *);

/* is_transu(C): if the compound C is an isotope of Np or Pu, return
   93 or 94 accordingly, otherwise return 0.
*/
int is_transu(struct cmpd *);

/* longevity(C): return, er, the longevity of the compound C. This
   assumes that C is 1 day old and atomic.
*/
int longevity(struct cmpd *);

/* kill_periodic_table(): free the memory used by the periodic table. */
void kill_periodic_table(void);

#endif
