#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "aa.h"
#include "mem.h"

#define GUY 24

static int g_print;
static struct list* g_interesting = NULL;
static struct list* g_ancestors = NULL;

/* It follows from the splitting theorem that a 3-day old string
   of the form L2R splits as L2.R iff R either starts with an isotope
   of a string from ctails or is an isotope of a string from ctailse.
   For L(1 or 3)R, the possibilities are obtained by prefixing 22
   to these strings, plus the case R = 22. However, L(1 or 3)22 is
   a complete N-day-old string iff L(1 or 3) is, so we can ignore
   that case. Also L(>3)R always splits.
*/
static const char *ctails[] = {
    "123", "131", "132", "141", "142",
    "143", "111", "312", "314", "3112",
    "3113", "3114","321", "3221", "3223", 
};

static const char *ctailse[] = {
    "13", "14", "3", "31", "311", "322",
};

#define NTAILS ((sizeof ctails) / (sizeof ctails[0]))
#define NTAILSE ((sizeof ctailse) / (sizeof ctailse[0]))
/* These arrays hold our internal versions of the strings mentioned
   in the previous comment.
*/
static struct cmpd *tails13[NTAILS];
static struct cmpd *tails13e[NTAILSE];
static struct cmpd *tails2[NTAILS];
static struct cmpd *tails2e[NTAILSE];

static void usage(const char *prog)
{
    fprintf(stderr, "usage: %s [-p] age [level]\n"
            " age >= 3, 1 <= level <= 3\n",
            prog); 
    exit(EXIT_FAILURE);
}

/* inittails(): set up the strings that can appear after splittings. */
static void inittails(void)
{
    int i;

    for (i = 0; i < NTAILS; ++i) {
        tails13[i] = readcmpd("22");
        tails13[i]->next->next = tails2[i] = readcmpd(ctails[i]);
    }
    for (i = 0; i < NTAILSE; ++i) {
        tails13e[i] = readcmpd("22");
        tails13e[i]->next->next = tails2e[i] = readcmpd(ctailse[i]);
    }
}

/* killtails(): free the memory allocated by inittails. */
static void killtails(void)
{
    int i;

    for (i = 0; i < NTAILS; ++i)
        killcmpd(tails13[i]);
    for (i = 0; i < NTAILSE; ++i) 
        killcmpd(tails13e[i]);
}

/* parent(C): C should be an even-length female string;
   the parent of C is returned.
*/
static struct cmpd *parent(const struct cmpd *c)
{
    struct cmpd h, *last;

    last = &h;
    while (c) {
        int i;

        assert(c->next);
        for (i = 0; i < c->n; ++i) {
            last->next = xalloc(sizeof *last);
            last = last->next;
            last->n = c->next->n;
        }
        c = c->next->next;
    }
    last->next = NULL;
    return h.next;
}

/* getmax(C): return the maximum of 3 and all the digits of C. */
static int getmax(struct cmpd *c)
{
    int ret = 3;

    while (c) {
        if (c->n > ret)
            ret = c->n;
        c = c->next;
    }
    return ret;
}

/* can_appear(C, N, AS, AE, S): return 1 if there is an N-day-old string
   of the form LCR, where L (resp R) must be empty if AS (resp AE) is
   non-zero. If S is non-zero, return 0 otherwise. If S is 0, false
   positives will sometimes be returned, at a considerable improvement
   in speed. Defined below.
*/
static int can_appear(struct cmpd *c, int age, int at_start, int at_end,
                      int strict);

/* can_be_parent(C, N, AS, AE): the same as can_appear(C, N, AS, AE, 1),
   with the added condition that the daughter of LCR must be the product
   of the daughters of L, C and R.
*/
static int can_be_parent(struct cmpd *c, int age, int at_start, int at_end)
{
    struct cmpd h, t, *e;
    int mh, mt;

    if (can_appear(c, age, 1, 1, 1))
        return 1;
    mh = getmax(c) + 1;
    if (!at_start) {
        h.next = c;
        for (h.n = 1; h.n <= mh; ++h.n) {
            if (h.n != c->n && can_appear(&h, age, 0, 1, 1))
                return 1;
        }
    }
    if (at_end)
        return 0;
    for (e = c; e->next; e = e->next)
        ;
    e->next = &t;
    t.next = NULL;
    mt = mh;
    for (t.n = 1; t.n <= mt; ++t.n) {
        if (t.n == e->n)
            continue;
        if (can_appear(c, age, 1, 0, 1)) {
            e->next = NULL;
            return 1;
        }
        if (!at_start) {
            if (t.n == mt)
                ++mh;
            for (h.n = 1; h.n <= mh; ++h.n) {
                if (h.n != c->n && can_appear(&h, age, 0, 0, 1)) {
                    e->next = NULL;
                    return 1;
                }
            }
        }
    }
    e->next = NULL;
    return 0;
}

/* can_be_girl(C, N, AS, AE, S): the same as can_appear(C, N, AS, AE, S),
   with the added condition that C must be a female substring of LCR.
*/
static int can_be_girl(struct cmpd *c, int age, int at_start, int at_end,
                       int strict)
{
    struct cmpd t, *e, *par;
    int even, max, ret;

    assert(age > 0);
    even = 1;
    max = 3;
    e = c;
    for (;;) {
        even = !even;
        if (e->n > max)
            max = e->n;
        if (!e->next)
            break;
        if (even && e->next->next && e->next->next->n == e->n)
            return 0;
        e = e->next;
    }
    if (!even) {
        if (at_end)
            return 0;
        e->next = &t;
        t.next = NULL;
        for (t.n = 1; t.n <= max + 1; ++t.n) {
            if (can_be_girl(c, age, at_start, 0, strict)) {
                e->next = NULL;
                return 1;
            }
        }
        e->next = NULL;
        return 0;
    }
    par = parent(c);
    if (strict)
        ret = can_be_parent(par, age - 1, at_start, at_end);
    else
        ret = can_appear(par, age - 1, at_start, at_end, 0);
    killcmpd(par);
    return ret;
}

/* can_appear(C, N, AS, AE, S): explanation above. */
static int can_appear(struct cmpd *c, int age, int at_start, int at_end,
                      int strict)
{
    struct cmpd h, *d;
    
    assert(c != NULL && age >= 0);
    if (age == 0) {
        if (strict && g_print) {
            struct list *temp = xalloc(sizeof *temp);
            temp->next = g_ancestors;
            temp->c = cpycmpd(c);
            g_ancestors = temp;
        }
        return 1;
    }
    if (can_be_girl(c, age, at_start, at_end, strict))
        return 1;
    if (at_start)
        return 0;
    d = c;
    while (d->next && d->next->next) {
        if (d->next->next->n == d->n)
            return 0;
        d = d->next->next;
    }
    h.next = c;
    for (h.n = 1; h.n <= 3; ++h.n) {
        if (can_be_girl(&h, age, 0, at_end, strict))
            return 1;
    }
    return 0;
}

/* can_split_before(C, N, AE, S): the same as can_appear(C, N, 0, AE, S),
   with the added condition that LCR must split as L.CR.
   Constraint: N >= 2 and, if AE == 0, C has the property that LCR splits
   as L.CR iff LC splits as L.C.
*/
static int can_split_before(struct cmpd *c, int age, int at_end, int strict)
{
    struct cmpd h;
    int max;

    if (can_appear(c, age, 1, at_end, strict))
        return 1;
    max = getmax(c);
    h.next = c;
    for (h.n = 1; h.n <= max + 1; ++h.n) {
        /* The constraint is needed to justify the use of cansplit(). */
        if (cansplit2(&h) && can_appear(&h, age, 0, at_end, strict))
            return 1;
    }
    return 0;
}

/* can_split_off(C, N, S): the same as can_appear(C, N, 0, 0, S),
   with the added condition that LCR must split as L.C.R.
   Constraint: N >= 3 and at most the last digit of C is > 3.
*/
static int can_split_off(struct cmpd *c, int age, int strict)
{
    struct cmpd **tails, **tailse, *e;
    int i, max;

    assert(c != NULL);
    max = getmax(c);
    for (e = c; e->next; e = e->next)
        ;
    if (e->n > 3) {
        /* Any string CR splits as C.R, so LCR splits as L.CR iff
           LC splits as L.C, and the constraint for can_split_before()
           is satisfied.
        */
        return can_split_before(c, age, 0, strict);
    }
    if (can_split_before(c, age, 1, strict))
        return 1;
    if (e->n == 2) {
        tails = tails2;
        tailse = tails2e;
    }
    else {
        tails = tails13;
        tailse = tails13e;
    }
    /* Since C has only digits <= 3, we do not need to consider isotopes
       of the strings in tails and tailse.
    */
    for (i = 0; i < NTAILS; ++i) {
        e->next = tails[i];
        /* The string being passed has the form CD and the property
           that CDR always splits as C.DR. It follows that LCDR splits
           as L.CDR iff LC splits as L.C, and the constraint for
           can_split_before() is satisfied.
        */
        if (can_split_before(c, age, 0, strict)) {
            e->next = NULL;
            return 1;
        }
    }
    for (i = 0; i < NTAILSE; ++i) {
        e->next = tailse[i];
        if (can_split_before(c, age, 1, strict)) {
            e->next = NULL;
            return 1;
        }
    }
    e->next = NULL;
    return 0;
}

/* is_interesting(C, N, L): if C is an interesting chunk of an N-day-old
   string, according to the criterion for being interesting at level L,
   return N + longevity(C); otherwise return 0.
*/
static int is_interesting(struct cmpd *c, int age, int level)
{
    int n;

    n = longevity(c) + age;
    if (n < GUY)
        return 0;
    if (level == 1)
        return n;
    if (!can_split_off(c, age, 0))
        return 0;
    if (level == 2)
        return n;
    if (!can_split_off(c, age, 1))
        return 0;
    return n;
}

/* nextlen(P, N, LEN, LEV): on entry, P contains all atomic strings of
   length LEN that can be chunks of N-day-old strings (and some that
   can't). Construct and return the corresponding list for length
   LEN + 1, freeing the memory used by P in the process. Also check
   for interesting strings according to level LEV, maintain the running
   counts, and report progress.
*/
static struct list *nextlen(struct list *prev, int age, int len, int level)
{
    struct list *ret = NULL,  *temp;
    struct cmpd c;
    int ct = 0, n;
    static int n_interesting = 0;
    static int cconst = 0;
    
    ++len;
    while (prev) {
        c.next = prev->c;
        for (c.n = 1; c.n <= 3; ++c.n) {
            if (!cansplit2(&c) && can_appear(&c, age, 0, 0, 0)) {
                temp = xalloc(sizeof *temp);
                temp->next = ret;
                ret = temp;
                ret->c = cpycmpd(&c);
                ++ct;
                if ((n = is_interesting(&c, age, level))) {
                    if (g_print) {
                        temp = xalloc(sizeof *temp);
                        temp->next = g_interesting;
                        g_interesting = temp;
                        g_interesting->c = cpycmpd(&c);
                    }
                    ++n_interesting;
                    if (n > cconst)
                        cconst = n;
                }
            }
        }
        temp = prev;
        prev = prev->next;
        killcmpd(temp->c);
        xfree(temp);
    }
    printf("At length %d: %d strings survive.\n", len, ct);
    printf("So far: interesting strings: %d;"
           " cosmological constant: %d.\n\n", n_interesting, cconst);
    fflush(stdout);
    return ret;
}

int main(int argc, char **argv)
{
    int len, age = 0, level = 0, i;
    struct list *p;
    char *prog;

    /* Process the arguments. */
    prog = argc > 0 && **argv ? argv[0] : "proof";
    if (argc == 0)
        usage(prog);
    g_print = 0;
    while (*++argv && !level) {
        if (**argv == '-') {
            if (strcmp(*argv, "-p") != 0)
                usage(prog);
            g_print = 1;
        }
        else if (!age) {
            age = atoi(*argv);
            if (age < 3)
                usage(prog);
        }
        else {
            level = atoi(*argv);
            if (level < 1 || level > 3)
                usage(prog);
        }
    }
    if (!age)
        usage(prog);
    if (!level)
        level = 1;
    if (*argv) {    
        if (strcmp(*argv, "-p") != 0)
            usage(prog);
        g_print = 1;
    }
    
    /* Initialise. */
    init_periodic_table();
    inittails();

    p = NULL;
    for (i = 4; i >= 1; --i) {
        struct list *temp = p;
        p = xalloc(sizeof *p);
        p->next = temp;
        p->c = xalloc(sizeof *p->c);
        p->c->n = i;
        p->c->next = NULL;
    }

    /* Hope this terminates. */
    for (len = 1; p; ++len)
        p = nextlen(p, age, len, level);

    /* Final messages and cleanup. */
    if (g_print) {
        printf("The interesting strings for age %d "
               "and level %d are:\n", age, level);
        prtlist_nls(g_interesting);
        killlist(g_interesting);
        if (g_ancestors) {
            puts("Their ancestors are:");
            prtlist_nls(g_ancestors);
            killlist(g_ancestors);
        }
    }
    
    killlist(p);
    killtails();
    kill_periodic_table();
    assert(numalloced() == 0);
    return 0;
}

