#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "aa.h"
#include "mem.h"

/* NUME(X): number of elements in the array X. */
#define NUME(x) ((sizeof (x)) / (sizeof (x)[0]))

#define NDAYS 10 /* The number of days to run our machines.       */
#define NCA 84   /* The number of initial strings for machine CA. */
#define NIN 40   /* The number of inputs for the machines.        */

enum {INF0, INF02, INF20, INF2}; /* Types of endings for the inputs. */
static const char *ends[] = {    /* The corresponding strings to     */
    "", "1", "11", "113"         /* tack on to the 10-day-old state. */
};

/* The initial strings for machine C. */
static const char *C0[] = {
    "", "1", "3", "11", "13", "31", "111", "113", "131", "133", "311",
    "1113", "1131", "1133", "1311", "1331", "3111", "3113",
    "11131", "11133", "11311", "11331", "13111", "13113", "13311", "31131",
    "31133",
    "111311", "111331", "113111", "113311", "311311", "311331",
    "1113111", "1113311", "3113111", "3113311",
};

static struct cmpd *CA0[NCA];      /* The initial strings for CA. */
static int inputs[NIN][NDAYS + 1]; /* The inputs for CA.          */

#define MAXATOMS 64      /* Size of the following array.       */
static struct le {
    int n;
    struct cmpd *c;
} atoms[MAXATOMS];       /* The potential 14-day exotic atoms. */
static int natoms = 0;   /* The number of them.                */

/* cat(A, B): catenate the digit string represented by the character
   string B onto the end of the digit string A and return the result.
*/
static struct cmpd *cat(struct cmpd *a, const char *b)
{
    struct cmpd *t;

    if (!a)
        return readcmpd(b);
    for (t = a; t->next; t = t->next)
        ;
    t->next = readcmpd(b);
    return a;
}

/* takes2(S): return non-zero if the {1,3}-string S can be followed
   by 2^1 in a mature string, else 0.
*/
static int takes2(const char *s)
{
    const char *t;

    t = strstr(s, "1311");
    if (t && strlen(t) == 4)
        return 0;
    t = strstr(s, "1331");
    return !(t && strlen(t) == 4);
}

/* takes222(S): return non-zero if the {1,3}-string S can be followed
   by 2^3 in a mature string, else 0.
*/
static int takes222(const char *s)
{
    int i;

    for (i = strlen(s) - 1; i >= 2 && s[i-2] != s[i]
             && (s[i] != '3' || s[i-1] != '3'); i -= 2)
        ;
    return i < 1 || (i == 1 && (s[0] != '3' || s[1] != '3'));
}

/* initstates(): set up the initial states for the machines. */
static void initstates(void) {
    int i, j;
    
    for (i = j = 0; i < NUME(C0); ++i) {
        if (takes2(C0[i]))
            CA0[j++] = cat(readcmpd(C0[i]), "2");
        CA0[j++] = cat(readcmpd(C0[i]), "22");
        if (takes222(C0[i]))
            CA0[j++] = cat(readcmpd(C0[i]), "222");
    }
    assert(j == NCA);
}

/* initin(): set up the input strings. */
static void initin(void)
{
    int m, n, ni, i;
    int ns[] = {1, 3, 4, 7};
    
    ni = 0;

    /* 2^n(02)^infty, 0 <= n <= 3 */
    for (n = 0; n <= 3; ++n) {
        for (i = 0; i < n; ++i)
            inputs[ni][i] = 2;
        for ( ; i < NDAYS; i += 2) {
            inputs[ni][i] = 0;
            inputs[ni][i+1] = 2;
        }
        inputs[ni++][NDAYS] = n % 2 == 0 ? INF02 : INF20;
    }

    /* 0^n 2^infty, 0 <= n <= 2 */
    for (n = 0; n <= 2; ++n) {
        for (i = 0; i < n; ++i)
            inputs[ni][i] = 0;
        for ( ; i < NDAYS; ++i)
            inputs[ni][i] = 2;
        inputs[ni++][NDAYS] = INF2;
    }

    /* 2^n 0 2^infty, 1 <= n <= 3 */
    for (n = 1; n <= 3; ++n) {
        for (i = 0; i < n; ++i)
            inputs[ni][i] = 2;
        inputs[ni][i++] = 0;
        for ( ; i < NDAYS; ++i)
            inputs[ni][i] = 2;
        inputs[ni++][NDAYS] = INF2;
    }

    /* 2^n 0^infty, 0 <= n <= 7 */
    for (n = 0; n <= 7; ++n) {
        for (i = 0; i < n; ++i)
            inputs[ni][i] = 2;
        for ( ; i < NDAYS; ++i)
            inputs[ni][i] = 0;
        inputs[ni++][NDAYS] = INF0;
    }

    /* 2^m 0 2^n 0^infty, 0 <= m <= 2, n = 1,3,4,7 */
    for (m = 0; m <= 2; ++m) {
        for (n = 0; n < NUME(ns); ++n) {
            for (i = 0; i < m; ++i)
                inputs[ni][i] = 2;
            inputs[ni][i++] = 0;
            for ( ; i < ns[n] + m + 1; ++i)
                inputs[ni][i] = 2;
            for ( ; i < NDAYS; ++i)
                inputs[ni][i] = 0;
            inputs[ni++][NDAYS] = INF0;
        }
    }

    /* 2^3 0 2^n 0^infty, n = 1,3 */
    for (n = 1; n <= 3; n += 2) {
        for (i = 0; i < 3; ++i)
            inputs[ni][i] = 2;
        inputs[ni][i++] = 0;
        for ( ; i < n + 4; ++i)
            inputs[ni][i] = 2;
        for ( ; i < NDAYS; ++i)
            inputs[ni][i] = 0;
        inputs[ni++][NDAYS] = INF0;
    }

    /* 0 2 0 2^n 0^infty, n= 1,3,4,7 */
    for (n = 0; n < NUME(ns); ++n) {
        inputs[ni][0] = 0;
        inputs[ni][1] = 2;
        inputs[ni][2] = 0;
        for (i = 3 ; i < ns[n] + 3; ++i)
            inputs[ni][i] = 2;
        for ( ; i < NDAYS; ++i)
            inputs[ni][i] = 0;
        inputs[ni++][NDAYS] = INF0;
    }

    /* 0^n 2^4 0 2^3 0^infty, 0 <= n <= 2 */
    for (n = 0; n <= 2; ++n) {
        for (i = 0; i < n; ++i)
            inputs[ni][i] = 0;
        for ( ; i < n + 4; ++i)
            inputs[ni][i] = 2;
        inputs[ni][i++] = 0;
        for ( ; i < n + 8; ++i)
            inputs[ni][i] = 2;
        for ( ; i < NDAYS; ++i)
            inputs[ni][i] = 0;
        inputs[ni++][NDAYS] = INF0;
    }

    /* 2 0 2 0 2^3 0^infty */
    inputs[ni][0] = 2;
    inputs[ni][1] = 0;
    inputs[ni][2] = 2;
    inputs[ni][3] = 0;
    for (i = 4 ; i < 7; ++i)
        inputs[ni][i] = 2;
    for ( ; i < NDAYS; ++i)
        inputs[ni][i] = 0;
    inputs[ni++][NDAYS] = INF0;
    
    assert(ni == NIN);
}

/* machCA(S, N): return the state of machine CA after receiving a single
   input N in state S, and destroy S.
*/
static struct cmpd *machCA(struct cmpd *state, int in)
{
    struct cmpd *t;

    t = dtr(state);
    killcmpd(state);
    state = t;
    if (in == 2) {
        while (t->next)
            t = t->next;
        t->next = xalloc(sizeof *t);
        t = t->next;
        t->n = 2;
        t->next = NULL;
    }
    return state;
}

/* runCA(S, IN): run the machine CA for 10 days with initial
   state S and input IN, tack on the appropriate ending, and record
   any exotic elements we've not seen before. S is destroyed.
*/
static void runCA(struct cmpd *s, int *in)
{
    int i, isnew;
    struct list *p, *q;
        
    for (i = 0; i < NDAYS; ++i)
        s = machCA(s, in[i]);
    p = split(cat(s, ends[in[NDAYS]]), 2);
    do {
        if (is_common(p->c))
            isnew = 0;
        else {
            for (i = 0; i < natoms
                     && cmpcmpd(atoms[i].c, p->c) != 0; ++i)
                ;
            isnew = i == natoms;
        }
        q = p;
        p = p->next;
        if (isnew) {
            assert(natoms < MAXATOMS);
            atoms[natoms].n = longevity(q->c);
            atoms[natoms].c = q->c;
            ++natoms;
        }
        else
            killcmpd(q->c);
        xfree(q);
    } while (p);
}

/* cmp_le(A, B): A and B are really of type struct le *; compare
   by lexicographic order of (longevity, string).
*/
static int cmp_le(const void *a, const void *b)
{
    const struct le *a1, *b1;

    a1 = (const struct le *)a;
    b1 = (const struct le *)b;
    return a1->n != b1->n ? a1->n - b1->n : cmpcmpd(a1->c, b1->c);
}
        
int main(void)
{
    int i, j;

    assert(NUME(C0) == 37);

    init_periodic_table();
    initstates();
    initin();

    for (i = 0; i < NCA; ++i)
        for (j = 0; j < NIN; ++j) 
            runCA(cpycmpd(CA0[i]), inputs[j]);

    qsort(atoms, natoms, sizeof atoms[0], cmp_le);
    puts("The potential 14-day-old exotic atoms, "
         "with their longevities, are:");
    for (i = 0; i < natoms; ++i) {
        printf("%2d: ", atoms[i].n);
        putcmpd(atoms[i].c);
        killcmpd(atoms[i].c);
    }
    printf("There are %d of them, with maximum longevity %d.\n",
           natoms, atoms[natoms - 1].n);
    
    for (i = 0; i < NCA; ++i)
        killcmpd(CA0[i]);
    kill_periodic_table();
    assert(numalloced() == 0);

    return 0;
}
